"""
Version and license information.
"""

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '3.0.3'
__versiondate__ = '2025-10-07'
__license__ = f'Starsim {__version__} ({__versiondate__}) — © 2023-2025 by IDM'
